@@ -8,14 +8,13 @@ from collections import deque
 from functools import wraps
 from inspect import isclass
 from inspect import isfunction
+from inspect import isgenerator
+from inspect import isgeneratorfunction
 from inspect import ismethod
 from inspect import ismethoddescriptor
 from inspect import isroutine
 from logging import getLogger
 
-from types import GeneratorType
-
-
 __all__ = 'weave', 'Aspect', 'Proceed', 'Return', 'ALL_METHODS', 'NORMAL_METHODS'
 
 try:
@@ -58,6 +57,26 @@ REGEX_TYPE = type(NORMAL_METHODS)
 VALID_IDENTIFIER = re.compile(r'^[^\W\d]\w*$', re.UNICODE if PY3 else 0)
 
 
+class UnacceptableAdvice(RuntimeError):
+    pass
+
+
+class ExpectedGenerator(TypeError):
+    pass
+
+
+class ExpectedGeneratorFunction(ExpectedGenerator):
+    pass
+
+
+class ExpectedAdvice(TypeError):
+    pass
+
+
+class UnsupportedType(TypeError):
+    pass
+
+
 class Proceed(object):
     """
     Instructs the Aspect Calls to call the decorated function. Can be used multiple times.
@@ -81,6 +100,16 @@ class Return(object):
         self.value = value
 
 
+class Yield(object):
+    """
+    Instructs the Aspect to yield a value.
+    """
+    __slots__ = 'value',
+
+    def __init__(self, value):
+        self.value = value
+
+
 class Aspect(object):
     """
     Container for the advice yielding generator. Can be used as a decorator on other function to change behavior
@@ -89,35 +118,121 @@ class Aspect(object):
     __slots__ = 'advise_function'
 
     def __init__(self, advise_function):
-        assert callable(advise_function)
+        if not isgeneratorfunction(advise_function):
+            raise ExpectedGeneratorFunction("advise_function %s must be a generator function." % advise_function)
         self.advise_function = advise_function
 
     def __call__(self, cutpoint_function):
-        @wraps(cutpoint_function)
-        def advice_wrapper(*args, **kwargs):
-            advisor = self.advise_function(*args, **kwargs)
-            if not isinstance(advisor, GeneratorType):
-                raise RuntimeError("advise_function %s did not return a generator." % self.advise_function)
-            advice = advisor.send(None)
-            while True:
-                if advice is Proceed or advice is None or isinstance(advice, Proceed):
-                    if advice is not Proceed:
-                        args = advice.args
-                        kwargs = advice.kwargs
+        if isgeneratorfunction(cutpoint_function):
+            if PY3:
+                return __import__(
+                    'aspectlib.py3support',
+                    fromlist=['decorate_advising_generator_py3']
+                ).decorate_advising_generator_py3(
+                    self.advise_function,
+                    cutpoint_function
+                )
+            else:
+                @wraps(cutpoint_function)
+                def advising_generator_wrapper(*args, **kwargs):
+                    advisor = self.advise_function(*args, **kwargs)
+                    if not isgenerator(advisor):
+                        raise ExpectedGenerator("advise_function %s did not return a generator." % self.advise_function)
                     try:
-                        result = cutpoint_function(*args, **kwargs)
-                    except Exception:
-                        advice = advisor.throw(*sys.exc_info())
-                    else:
-                        try:
-                            advice = advisor.send(result)
-                        except StopIteration:
-                            return result
-                elif advice is Return:
-                    return
-                elif isinstance(advice, Return):
-                    return advice.value
-        return advice_wrapper
+                        advice = next(advisor)
+                        while True:
+                            logger.debug('Got advice %r from %s', advice, self.advise_function)
+                            if advice is Proceed or advice is None or isinstance(advice, Proceed):
+                                if isinstance(advice, Proceed):
+                                    args = advice.args
+                                    kwargs = advice.kwargs
+                                gen = cutpoint_function(*args, **kwargs)
+                                try:
+                                    try:
+                                        generated = next(gen)
+                                    except StopIteration as exc:
+                                        logger.exception("The cutpoint has been exhausted (early).")
+                                        result = exc.args and exc.args[0]
+                                    else:
+                                        while True:
+                                            try:
+                                                sent = yield generated
+                                            except GeneratorExit as exc:
+                                                logger.exception("Got GeneratorExit while consuming the cutpoint")
+                                                gen.close()
+                                                raise exc
+                                            except BaseException as exc:
+                                                logger.exception("Got exception %r. Throwing it the cutpoint", exc)
+                                                try:
+                                                    generated = gen.throw(*sys.exc_info())
+                                                except StopIteration as exc:
+                                                    logger.exception("The cutpoint has been exhausted.")
+                                                    result = exc.args and exc.args[0]
+                                                    break
+                                            else:
+                                                try:
+                                                    if sent is None:
+                                                        generated = next(gen)
+                                                    else:
+                                                        generated = gen.send(sent)
+                                                except StopIteration as exc:
+                                                    logger.exception("The cutpoint has been exhausted.")
+                                                    result = exc.args and exc.args[0]
+                                                    break
+                                except BaseException as exc:
+                                    advice = advisor.throw(*sys.exc_info())
+                                else:
+                                    try:
+                                        advice = advisor.send(result)
+                                    except StopIteration:
+                                        return
+                                finally:
+                                    gen.close()
+                            elif advice is Return:
+                                return
+                            elif isinstance(advice, Return):
+                                raise StopIteration(advice.value)
+                            elif isinstance(advice, Yield):
+                                item = yield advice.value
+                                logger.critical('sending %s (from yield %r)', item, advice.value)
+                                advice = advisor.send(item)
+                            else:
+                                raise UnacceptableAdvice("Unknown advice %s" % advice)
+                    finally:
+                        advisor.close()
+                return advising_generator_wrapper
+        else:
+            @wraps(cutpoint_function)
+            def advising_function_wrapper(*args, **kwargs):
+                advisor = self.advise_function(*args, **kwargs)
+                if not isgenerator(advisor):
+                    raise ExpectedGenerator("advise_function %s did not return a generator." % self.advise_function)
+                try:
+                    advice = next(advisor)
+                    while True:
+                        logger.debug('Got advice %r from %s', advice, self.advise_function)
+                        if advice is Proceed or advice is None or isinstance(advice, Proceed):
+                            if isinstance(advice, Proceed):
+                                args = advice.args
+                                kwargs = advice.kwargs
+                            try:
+                                result = cutpoint_function(*args, **kwargs)
+                            except Exception:
+                                advice = advisor.throw(*sys.exc_info())
+                            else:
+                                try:
+                                    advice = advisor.send(result)
+                                except StopIteration:
+                                    return result
+                        elif advice is Return:
+                            return
+                        elif isinstance(advice, Return):
+                            return advice.value
+                        else:
+                            raise UnacceptableAdvice("Unknown advice %s" % advice)
+                finally:
+                    advisor.close()
+            return advising_function_wrapper
 
 
 class Fabric(object):
@@ -206,10 +321,10 @@ def weave(target, aspects, **options):
     """
     if not callable(aspects):
         if not hasattr(aspects, '__iter__'):
-            raise RuntimeError('%s must be an `Aspect` instance, a callable or an iterable of.' % aspects)
+            raise ExpectedAdvice('%s must be an `Aspect` instance, a callable or an iterable of.' % aspects)
         for obj in aspects:
             if not callable(obj):
-                raise RuntimeError('%s must be an `Aspect` instance or a callable.' % obj)
+                raise ExpectedAdvice('%s must be an `Aspect` instance or a callable.' % obj)
     assert target, "Can't weave falsy value %r." % target
     if isinstance(target, (list, tuple)):
         return Rollback([
@@ -290,7 +405,7 @@ def weave(target, aspects, **options):
     elif isclass(target):
         return weave_class(target, aspects, **options)
     else:
-        raise RuntimeError("Can't weave object %s of type %s" % (target, type(target)))
+        raise UnsupportedType("Can't weave object %s of type %s" % (target, type(target)))
 
 
 def rewrap_method(func, klass, aspect):
