@@ -205,13 +205,42 @@ class SlotsTestSubSubClass(SlotsTestSubClass):
 def test_aspect_bad():
     @aspectlib.Aspect
     def aspect():
+        yield
+
+    def aspect_fail():
         return "crap"
 
+    aspect.advise_function = aspect_fail
+
     @aspect
     def func():
         pass
 
-    raises(RuntimeError, func)
+    raises(aspectlib.ExpectedGenerator, func)
+
+
+def test_aspect_bad_gen():
+    @aspectlib.Aspect
+    def aspect():
+        yield
+
+    def aspect_fail():
+        return "crap"
+
+    aspect.advise_function = aspect_fail
+
+    @aspect
+    def func():
+        yield
+
+    raises(aspectlib.ExpectedGenerator, list, func())
+
+
+def test_aspect_bad_decorate():
+    def aspect():
+        return "crap"
+
+    raises(aspectlib.ExpectedGeneratorFunction, aspectlib.Aspect, aspect)
 
 
 def test_aspect_return():
@@ -261,6 +290,7 @@ def test_aspect_raise_from_aspect():
     @aspectlib.Aspect
     def aspect():
         1/0
+        yield
 
     @aspect
     def func():
@@ -295,11 +325,11 @@ def test_weave_func():
 
 
 def test_broken_aspect():
-    raises(RuntimeError, aspectlib.weave, None, None)
+    raises(aspectlib.ExpectedAdvice, aspectlib.weave, None, None)
 
 
 def test_weave_empty_target():
-    raises(RuntimeError, aspectlib.weave, (), None)
+    raises(aspectlib.ExpectedAdvice, aspectlib.weave, (), None)
 
 
 def test_weave_missing_global(cls=Global):
@@ -997,7 +1027,7 @@ def test_weave_unknown():
     def aspect():
         yield aspectlib.Proceed
 
-    raises(RuntimeError, aspectlib.weave, 1, aspect)
+    raises(aspectlib.UnsupportedType, aspectlib.weave, 1, aspect)
 
 
 def test_weave_unimportable():
@@ -1096,8 +1126,303 @@ def test_list_of_aspects():
         raises(TypeError, module_func, 1, 2, 3)
         assert module_func.calls == [(None, (1, 2, 3), {})]
 
+
 def test_list_of_invalid_aspects():
     raises(AssertionError, aspectlib.weave, module_func, [lambda func: None])
     raises(TypeError, aspectlib.weave, module_func, [lambda: None])
-    raises(RuntimeError, aspectlib.weave, module_func, [None])
-    raises(RuntimeError, aspectlib.weave, module_func, ['foobar'])
+    raises(aspectlib.ExpectedAdvice, aspectlib.weave, module_func, [None])
+    raises(aspectlib.ExpectedAdvice, aspectlib.weave, module_func, ['foobar'])
+
+
+def test_aspect_on_func():
+    hist = []
+
+    @aspectlib.Aspect
+    def aspect():
+        try:
+            hist.append('before')
+            hist.append((yield aspectlib.Proceed))
+            hist.append('after')
+        except Exception:
+            hist.append('error')
+        finally:
+            hist.append('finally')
+        try:
+            hist.append((yield aspectlib.Return('squelched')))
+        except GeneratorExit:
+            hist.append('closed')
+            raise
+        else:
+            hist.append('consumed')
+
+    @aspect
+    def func():
+        raise RuntimeError()
+    assert func() == 'squelched'
+    assert hist == ['before', 'error', 'finally', 'closed']
+
+
+def test_aspect_on_func_invalid_advice():
+    hist = []
+
+    @aspectlib.Aspect
+    def aspect():
+        yield aspectlib.Yield("stuff")
+
+    @aspect
+    def func():
+        raise RuntimeError()
+
+    raises(aspectlib.UnacceptableAdvice, func)
+
+def test_aspect_on_generator_func():
+    hist = []
+
+    @aspectlib.Aspect
+    def aspect():
+        try:
+            hist.append('before')
+            yield aspectlib.Yield('prefix')
+            yield aspectlib.Proceed
+            yield aspectlib.Yield('bad-suffix')
+            hist.append('after')
+        except Exception:
+            hist.append('error')
+        finally:
+            hist.append('finally')
+        yield aspectlib.Yield('suffix')
+        try:
+            hist.append((yield aspectlib.Return))
+        except GeneratorExit:
+            hist.append('closed')
+            raise
+        else:
+            hist.append('consumed')
+        hist.append((yield aspectlib.Yield('very-bad-suffix')))
+
+    @aspect
+    def func():
+        for i in range(3):
+            yield i
+        raise RuntimeError()
+    assert list(func()) == ['prefix', 0, 1, 2, 'suffix']
+    print (hist)
+    assert hist == ['before', 'error', 'finally', 'closed']
+
+
+def test_aspect_on_generator_func_bad_advice():
+    @aspectlib.Aspect
+    def aspect():
+        yield 'crappo'
+
+    @aspect
+    def func():
+        for i in range(3):
+            yield i
+        raise RuntimeError()
+
+    raises(aspectlib.UnacceptableAdvice, list, func())
+
+
+def test_aspect_on_generator_different_args():
+    @aspectlib.Aspect
+    def aspect():
+        yield aspectlib.Proceed('something')
+
+    @aspect
+    def func(arg=None):
+        yield arg
+
+    assert list(func()) == ['something']
+
+
+def test_aspect_on_generator_raise_stopiteration():
+    @aspectlib.Aspect
+    def aspect():
+        val = yield aspectlib.Proceed
+        yield aspectlib.Yield(val)
+
+    @aspect
+    def func():
+        raise StopIteration('something')
+        yield
+
+    assert list(func()) == ['something']
+
+
+def test_aspect_on_generator_close():
+    excs = []
+    @aspectlib.Aspect
+    def aspect():
+        yield aspectlib.Proceed
+
+    @aspect
+    def func():
+        try:
+            yield 'something'
+        except BaseException as exc:
+            excs.append(type(exc))
+
+    assert list(func()) == ['something']
+    assert excs == []
+
+    gen = func()
+    next(gen)
+    gen.close()
+    del gen
+    assert excs == [GeneratorExit]
+
+
+def test_aspect_on_generator_throw():
+    excs = []
+    @aspectlib.Aspect
+    def aspect():
+        yield aspectlib.Proceed
+
+    @aspect
+    def func():
+        try:
+            yield 'something'
+        except BaseException as exc:
+            excs.append(type(exc))
+        yield 'lastthing'
+
+    assert list(func()) == ['something', 'lastthing']
+    assert excs == []
+
+    gen = func()
+    print(next(gen))
+    gen.throw(RuntimeError)
+    assert excs == [RuntimeError]
+
+
+def test_aspect_on_generator_throw_exhaust():
+    excs = []
+    @aspectlib.Aspect
+    def aspect():
+        yield aspectlib.Proceed
+
+    @aspect
+    def func():
+        try:
+            yield 'something'
+        except BaseException as exc:
+            excs.append(type(exc))
+
+    assert list(func()) == ['something']
+    assert excs == []
+
+    gen = func()
+    print(next(gen))
+    raises(StopIteration, gen.throw, RuntimeError)
+    assert excs == [RuntimeError]
+
+
+def test_aspect_on_generator_send_in_aspect():
+    a_values = []
+    f_values = []
+
+    @aspectlib.Aspect
+    def aspect():
+        a_values.append((yield aspectlib.Yield('value')))
+        yield aspectlib.Proceed
+
+    @aspect
+    def func():
+        f_values.append((yield 'something'))
+        yield
+        yield
+        yield
+
+    gen = func()
+    gen.send(None)
+    gen.send(1)
+    gen.send(2)
+
+    assert a_values == [1]
+    assert f_values == [2]
+
+
+def test_aspect_on_generator_result_from_aspect():
+    @aspectlib.Aspect
+    def aspect():
+        yield aspectlib.Proceed
+        yield aspectlib.Return('result')
+
+    @aspect
+    def func():
+        yield 'something'
+
+    gen = func()
+    try:
+        while 1:
+            next(gen)
+    except StopIteration as exc:
+        assert exc.args == ('result',)
+    else:
+        raise AssertionError("did not raise StopIteration")
+
+
+def test_aspect_on_generator_result():
+    result = []
+    @aspectlib.Aspect
+    def aspect():
+        result.append((yield aspectlib.Proceed))
+
+    @aspect
+    def func():
+        yield 'something'
+        raise StopIteration('value')
+
+    assert list(func()) == ['something']
+    assert result == ['value']
+
+
+def test_aspect_on_coroutine():
+    hist = []
+
+    @aspectlib.Aspect
+    def aspect():
+        try:
+            hist.append('before')
+            hist.append((yield aspectlib.Yield('prefix')))
+            hist.append((yield aspectlib.Proceed))
+            hist.append((yield aspectlib.Yield('suffix')))
+            hist.append('after')
+        except Exception:
+            hist.append('error')
+        finally:
+            hist.append('finally')
+        try:
+            hist.append((yield aspectlib.Return))
+        except GeneratorExit:
+            hist.append('closed')
+            raise
+        else:
+            hist.append('consumed')
+        hist.append((yield aspectlib.Yield('bad-suffix')))
+
+    @aspect
+    def func():
+        val = 99
+        for _ in range(3):
+            print("YIELD", val + 1)
+            val = yield val + 1
+            print("GOT", val)
+        raise StopIteration("the-return-value")
+    gen = func()
+    data = []
+    try:
+        for i in [None, 'captured', 0, 1, 'last-one', 'captured', 2, 3, 4, 5, 6, 7, 8, 9]:
+            data.append(gen.send(i))
+    except StopIteration:
+        data.append('done')
+    print(data)
+    assert data == ['prefix', 100, 1, 2, 'suffix', 'done']
+    print (hist)
+    assert hist == ['before', 'captured', 'the-return-value', 'captured', 'after', 'finally', 'closed']
+
+if __name__ == '__main__':
+    test_aspect_on_generator_func_bad_advice()
+    import os
+    os._exit(1)
